"""
Special arguments parser
"""

import regex

SPECIAL_PARSER = regex.compile(r"(?P<special>\D+)(-?\d*)", regex.IGNORECASE)

strings = {"last": ["l", "latest"], "last*": ["l*", "latest*", "latest-all"]}


def get_qualified_name(special):
    for key, values in strings.items():
        if special.lower() in (*values, key):
            return key


def special_parser(streams, string):

    returnee = []

    for match in SPECIAL_PARSER.finditer(string):
        special, index = match.group("special", 2)
        index = None if not index else int(index)

        special = get_qualified_name(special)

        if special is None:
            continue

        if special == "last":
            returnee.extend([streams[-(index or 1)]])
        else:
            if special == "last*":
                returnee.extend(streams[-(index or 1) :])
                returnee.extend(streams[: -(index or 1)])
            """
            More specials to be thought about.
            """

    yield from returnee



from animdl2.utils import optopt

ENDPOINT = "https://api.allanime.day/api"

ALLANIME_SEARCH_GQL = """\
query(
        $search: SearchInput
        $limit: Int
        $page: Int
        $translationType: VaildTranslationTypeEnumType
        $countryOrigin: VaildCountryOriginEnumType
    ) {
    shows(
        search: $search
        limit: $limit
        page: $page
        translationType: $translationType
        countryOrigin: $countryOrigin
    ) {
        pageInfo {
            total
        }
        edges {
            %s
        }
    }
}"""


ALLANIME_EPISODES_GQL = """\
query ($showId: String!, $translationType: VaildTranslationTypeEnumType!, $episodeString: String!) {
    episode(
        showId: $showId
        translationType: $translationType
        episodeString: $episodeString
    ) {
        %s
    }
}"""

ALLANIME_SHOW_GQL = """
query ($showId: String!) {
    show(
        _id: $showId
    ) {
        %s
    }
}
"""


class AllAnimeGQLAPI:
    """
    Fast and effective GQL API wrapper for AllAnime.
    """

    info_table = {}

    def __init__(self, *, endpoint: str = ENDPOINT):
        self.api_endpoint = endpoint

    def fetch_gql(
        self,
        session,
        query: str,
        variables: dict = None,
        *,
        persistent_query: str = None,
    ):
        variables = optopt.jsonlib.dumps(variables)

        response = session.get(
            self.api_endpoint,
            params={
                "variables": variables,
                "query": query,
            },
            headers={
                "Referer": "https://allanime.to/",
            },
        ).json()

        if response.get("errors"):
            if persistent_query is not None:
                response = session.get(
                    self.api_endpoint,
                    params={
                        "variables": variables,
                        "extensions": optopt.jsonlib.dumps(
                            {
                                "persistedQuery": {
                                    "version": 1,
                                    "sha256Hash": persistent_query,
                                }
                            }
                        ),
                    },
                ).json()

                if response.get("errors"):
                    return {}

                return response.get("data", {})

            return {}

        return response.get("data", {})

    def iter_search_results(
        self,
        session,
        query: str,
        translation_type=None,
        country_origin=None,
        show_nsfw=True,
        show_unknown=True,
        limit=None,
        *,
        keys: tuple = (
            "_id",
            "name",
            "availableEpisodesDetail",
        ),
    ):
        search = {
            "allowAdult": show_nsfw,
            "allowUnknown": show_unknown,
            "query": query,
        }

        variables = {
            "search": search,
        }

        if translation_type is not None:
            variables["translationType"] = translation_type

        if country_origin is not None:
            variables["countryOrigin"] = country_origin

        current_count = 0
        total = None
        page = 1

        gql = ALLANIME_SEARCH_GQL % " ".join(keys)

        while limit is None or current_count < limit:
            variables["page"] = page
            variables["limit"] = limit or 40

            response = self.fetch_gql(session, gql, variables)

            edges = response.get("shows", {}).get("edges", [])

            if not edges:
                return

            for data in edges:
                self.info_table.setdefault(data["_id"], {}).update(data)

                yield data

            current_count += len(edges)
            page += 1

            total = total or (
                response.get("shows", {}).get("pageInfo", {}).get("total", limit)
            )
            if limit is None:
                limit = total

    def fetch_show_info(
        self,
        session,
        show_id: str,
        keys: tuple = (
            "_id",
            "name",
            "availableEpisodesDetail",
        ),
    ):
        if show_id in self.info_table and (
            keys is None or all(key in self.info_table[show_id] for key in keys)
        ):
            return self.info_table[show_id]

        variables = {
            "showId": show_id,
        }

        response = self.fetch_gql(
            session,
            ALLANIME_SHOW_GQL % " ".join(keys),
            variables,
        )

        self.info_table.setdefault(show_id, {}).update(response.get("show", {}))

        return response.get("show", {})

    def fetch_episode(
        self,
        session,
        show_id: str,
        episode_string: str,
        *,
        translation_type: str = "sub",
        keys: tuple = ("episodeString", "sourceUrls", "notes"),
    ):
        variables = {
            "showId": show_id,
            "translationType": translation_type,
            "episodeString": episode_string,
        }

        return self.fetch_gql(
            session,
            ALLANIME_EPISODES_GQL % " ".join(keys),
            variables,
            persistent_query="0ac09728ee9d556967c1a60bbcf55a9f58b4112006d09a258356aeafe1c33889",
        )


api = AllAnimeGQLAPI()



from functools import partial

import lxml.html as htmlparser
import regex

from ....config import NINEANIME
from ...helpers import construct_site_based_regex
from .decipher import vrf_api

CONTENT_ID_REGEX = regex.compile(r'data-id="(.+?)"')

REGEX = construct_site_based_regex(
    NINEANIME, extra_regex=r"/watch/[^&?/]+\.(?P<slug>[^&?/]+)"
)
TITLES_REGEX = regex.compile(r'<h1 .+? class="title d-title" .+?>(.+?)</h1>')

SOURCES = {
    "41": "vidstream",
    "28": "mycloud",
    "35": "mp4upload",
    "40": "streamtape",
    "43": "videovard",
    "44": "filemoon",
}


def fetch_episode(session, data_source):
    episode_ids = data_source.get("data-ids")

    response = htmlparser.fromstring(
        session.get(
            NINEANIME + f"ajax/server/list/{episode_ids}",
            params={"vrf": vrf_api(session, episode_ids)},
        ).json()["result"]
    )

    for content_type_container in response.cssselect("div[data-type]"):
        for servers in content_type_container.cssselect("ul > li[data-sv-id]"):
            source = SOURCES[servers.get("data-sv-id")]
            link_id = servers.get("data-link-id")

            enc_content_url = session.get(
                NINEANIME + f"ajax/server/{link_id}",
                params={"vrf": vrf_api(session, link_id)},
            ).json()["result"]["url"]

            content_url = vrf_api(session, enc_content_url, endpoint="decrypt")

            yield {
                "stream_url": content_url,
                "further_extraction": (
                    source,
                    {"headers": {"referer": NINEANIME}},
                ),
                "title": f"Episode {data_source.get('data-num')}, {content_type_container.get('data-type').upper()}",
            }


def fetcher(session, url, check, match):
    content_id = CONTENT_ID_REGEX.search(session.get(url).text).group(1)

    for data_source in htmlparser.fromstring(
        session.get(
            NINEANIME + f"ajax/episode/list/{content_id}",
            params={"vrf": vrf_api(session, content_id)},
        ).json()["result"]
    ).cssselect("a[data-num]"):
        episode_string: str = data_source.get("data-num")

        episode = int(episode_string) if episode_string.isdigit() else 0

        if check(episode):
            yield partial(
                lambda data_source: list(fetch_episode(session, data_source)),
                data_source,
            ), episode


def metadata_fetcher(session, url, match):
    response = session.get(url).text

    return {"titles": TITLES_REGEX.findall(response)}



"""
async_downloader.py but one that utilises threads.
"""


import threading
import time
from typing import TYPE_CHECKING

import httpx

if TYPE_CHECKING:
    from io import FileIO
    from typing import Optional

    from tqdm import tqdm


class MultiConnectionDownloader:
    MINIMUM_PART_SIZE = 1024**2
    MAX_RETRIES = 5

    def __init__(
        self,
        session: httpx.Client,
        *args,
        progress_bar: "Optional[tqdm]" = None,
        retry_timeout: "Optional[int]" = None,
        **kwargs,
    ):
        self.session = session
        self.progress_bar = progress_bar
        self.retry_timeout = retry_timeout

        self.args = args
        self.kwargs = kwargs

        self.io_lock = threading.Lock()
        self.active_threads: "dict[str, threading.Thread]" = {}
        self.threads_errors: "dict[str, Exception]" = {}

    def download_section(
        self,
        io: "FileIO",
        start: int,
        end: int,
        progress_bar: "Optional[tqdm]" = None,
        pause_event: "Optional[threading.Event]" = None,
        error_event: "Optional[threading.Event]" = None,
    ):
        kwargs = self.kwargs.copy()

        headers = kwargs.pop("headers") or {}
        content_length = end
        position = start or 0

        is_incomplete = lambda: content_length is None or position < content_length
        is_downloading = lambda: (error_event is None or not error_event.is_set()) and (
            pause_event is None or not pause_event.is_set()
        )

        retry_count = 0

        while is_downloading() and is_incomplete():
            if content_length is None:
                if start is not None:
                    headers["Range"] = f"bytes={position}-"
            else:
                headers["Range"] = f"bytes={position}-{content_length}"

            try:
                with self.session.stream(
                    *self.args, **kwargs, headers=headers
                ) as response:
                    content_length = (
                        int(response.headers.get("Content-Length", 0)) or None
                    )

                    if progress_bar is not None:
                        if content_length > 0:
                            progress_bar.total = content_length

                    for chunk in response.iter_bytes(8192):
                        chunk_size = len(chunk)

                        if self.progress_bar is not None:
                            self.progress_bar.update(chunk_size)

                        if progress_bar is not None:
                            progress_bar.update(chunk_size)

                        self.write_to_file(
                            self.io_lock,
                            io,
                            position,
                            chunk,
                        )
                        position += chunk_size

                        if not is_downloading():
                            break

                    if content_length is None:
                        content_length = position

            except httpx.HTTPError as error:
                if retry_count >= self.MAX_RETRIES:
                    locks = ()

                    if progress_bar is not None:
                        locks += (progress_bar.get_lock(),)
                    if self.progress_bar is not None:
                        locks += (self.progress_bar.get_lock(),)

                    self.threads_errors[threading.current_thread().name] = error

                    if error_event is not None:
                        error_event.set()
                else:
                    if self.retry_timeout is not None:
                        time.sleep(self.retry_timeout)

                    retry_count += 1

        return (start, position)

    @staticmethod
    def write_to_file(
        lock: threading.Lock,
        io: "FileIO",
        position: int,
        data: bytes,
    ):
        with lock:
            io.seek(position)
            io.write(data)
            io.flush()

    def allocate_downloads(
        self,
        io: "FileIO",
        content_length: int = None,
        connections: int = 8,
        allocate_content_on_disk: bool = False,
        pause_event: "Optional[threading.Event]" = None,
        *,
        start_at: int = 0,
        end_at: int = None,
        threaded: bool = False,
    ):
        error_event = threading.Event()

        if end_at is not None:
            if content_length is None:
                raise ValueError("Cannot specify end_at when content_length is None")
            else:
                content_length = min(content_length, end_at)

        def iter_allocations():
            if not threaded or (
                content_length is None or content_length < self.MINIMUM_PART_SIZE
            ):
                yield None, None
            else:
                chunk_size = (content_length - start_at) // connections
                for i in range(connections - 1):
                    yield start_at + i * chunk_size, start_at + (i + 1) * chunk_size - 1

                yield start_at + (connections - 1) * chunk_size, None

        if allocate_content_on_disk:
            with self.io_lock:
                io.truncate(content_length)

        for n, (start, end) in enumerate(iter_allocations()):
            thread = threading.Thread(
                target=self.download_section,
                args=(io, start, end),
                kwargs={"pause_event": pause_event, "error_event": error_event},
                name=f"{self.__class__.__name__}@0x{id(self):x}-{n}",
            )
            thread.start()
            self.active_threads[thread.name] = thread

        for _, thread in self.active_threads.items():
            thread.join()


def prefetch(
    session: httpx.Client,
    *args,
    **kwargs,
):
    headers = kwargs.pop("headers") or {}
    headers["Range"] = "bytes=0-0"

    with session.stream(*args, **kwargs, headers=headers) as response:
        response.close()
        return response



import enum
import logging
import traceback
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Dict, Generator, Optional, Tuple
    import pathlib

    import httpx

from ...codebase import downloader, extractors
from . import (
    aniskip,
    banner,
    constants,
    decorators,
    intelliq,
    player,
    prompts,
    special,
    stream_handlers,
)
from .processors import process_query, provider_searcher_mapping

further_extraction_logger = logging.getLogger("further-extraction")


def further_extraction(session: "httpx.Client", parent_stream: "Dict[str, Any]"):
    extractor, options = parent_stream.pop("further_extraction", (None, None))

    if extractor is None:
        return (yield parent_stream)

    for ext_module, ext in extractors.iter_extractors():
        if ext != extractor or getattr(ext_module.extract, "disabled", False):
            continue

        try:
            for stream in ext_module.extract(
                session, parent_stream.get("stream_url"), **options
            ):
                child = parent_stream.copy()
                child.update(stream)

                yield from further_extraction(session, child)

        except Exception as exception:
            further_extraction_logger.error(
                f"Extraction from {ext!r} failed due to: {exception!r}."
            )


def ensure_extraction(session, stream_uri_caller):
    for stream in stream_uri_caller():
        if "further_extraction" in stream:
            yield from further_extraction(session, stream)
        else:
            yield stream


class SafeCaseEnum(enum.Enum):
    NO_CONTENT_FOUND = enum.auto()
    DOWNLOADED = enum.auto()

    EXTRACTION_ERROR = enum.auto()
    DOWNLOADER_EXCEPTION = enum.auto()


def safe_download_callback(
    session: "httpx.Client",
    logger: "logging.Logger",
    stream_urls: "Generator[Dict[str, Any], None, None]",
    quality: str,
    expected_download_path: "pathlib.Path",
    use_internet_download_manager: bool = False,
    retry_timeout: "Optional[int]" = None,
    log_level: "Optional[int]" = None,
    **kwargs,
) -> "Tuple[SafeCaseEnum, Optional[BaseException]]":
    flattened_streams = list(stream_urls)

    try:
        streams = intelliq.filter_quality(flattened_streams, quality)
    except Exception as exception:
        return SafeCaseEnum.EXTRACTION_ERROR, exception

    if not streams:
        return SafeCaseEnum.NO_CONTENT_FOUND, None

    for stream in streams:
        needs_further_extraction = "further_extraction" in stream

        if needs_further_extraction:
            status, potential_error = safe_download_callback(
                session,
                logger,
                further_extraction(session, stream),
                quality,
                expected_download_path,
                use_internet_download_manager,
                retry_timeout,
                log_level,
                **kwargs,
            )

            if status == SafeCaseEnum.DOWNLOADED:
                return status, potential_error

            if status == SafeCaseEnum.NO_CONTENT_FOUND:
                logger.debug(f"Could not find streams on further extraction, skipping.")
                continue

            if status == SafeCaseEnum.EXTRACTION_ERROR:
                logger.error(
                    f"Could not extract streams from further extraction due to an error: {potential_error}, skipping."
                )
                continue

            if status == SafeCaseEnum.DOWNLOADER_EXCEPTION:
                logger.error(
                    f"Could not download streams from further extraction due to multiple errors, skipping."
                )
                continue

        else:
            try:
                downloader.handle_download(
                    session=session,
                    url=stream["stream_url"],
                    expected_download_path=expected_download_path,
                    use_internet_download_manager=use_internet_download_manager,
                    headers=stream.get("headers"),
                    preferred_quality=quality,
                    subtitles=stream.get("subtitle"),
                    **kwargs,
                )
                return SafeCaseEnum.DOWNLOADED, None

            except Exception as download_exception:
                logger.error(
                    f"Could not download stream due to an error: {download_exception!r}, skipping."
                )
                logger.debug(f"Traceback for the above error: {traceback.format_exc()}")

        logger.critical(
            "Could not download any streams. Use the project with 0 log level to view errors for debugging and bug reporting purposes."
        )

        return SafeCaseEnum.NO_CONTENT_FOUND, None



import functools

import regex

from ...providers.nineanime.decipher import vrf_futoken

EMBED_URL_REGEX = regex.compile(r"(.+?)/(?:e(?:mbed)?)/([a-zA-Z0-9]+)\?t=(.+?)(?:&|$)")


@functools.lru_cache()
def futoken_resolve(session, url):
    return session.get(url).text


def extract(session, url, **opts):
    match = EMBED_URL_REGEX.search(url)
    host = match.group(1)

    futoken = futoken_resolve(session, host + "/futoken")
    data = vrf_futoken(
        session,
        futoken,
        "rawvizcloud" if host != "https://mcloud.to" else "rawmcloud",
        match.group(2),
    )
    vidstream_info = session.get(
        data + f"?t={match.group(3)}",
        headers={"referer": url},
    )

    return [
        {"stream_url": content.get("file", ""), "headers": {"referer": url}}
        for content in vidstream_info.json()["result"].get("sources", [])
    ]



import json
from functools import lru_cache

import regex
import yarl

from .utils import decipher_salted_aes

CONTENT_ID_REGEX = regex.compile(r"embed-6/([^?#&/.]+)")

SALT_SECRET_ENDPOINT = "https://github.com/enimax-anime/key/raw/e6/key.txt"


@lru_cache()
def get_associative_key(session, endpoint):
    return session.get(endpoint).json()


def extract(session, url, **opts):
    url = yarl.URL(url)

    ajax_response = session.get(
        f"https://{url.host}/embed-2/ajax/e-1/getSources",
        params={"id": url.name},
    )

    sources = ajax_response.json()

    key_finders = None
    encrypted: bool = sources["encrypted"]

    if encrypted:
        key_finders = get_associative_key(session, SALT_SECRET_ENDPOINT)

    subtitles = [
        _.get("file") for _ in sources.get("tracks") if _.get("kind") == "captions"
    ]

    if encrypted:
        retval = json.loads(decipher_salted_aes(sources["sources"], key_finders))

    else:
        retval = sources["sources"]

    def yielder():
        for _ in retval:
            yield {
                "stream_url": _["file"],
                "subtitle": subtitles,
            }

    return list(yielder())



import functools
import importlib
import inspect
import pathlib

DO_NOTHING = lambda *args, **kwargs: True


class ProjectContext(dict):
    """
    Context for the project so that we can set
    and get variables across various different
    activities in the project.

    (This is just a hashable dict.)

    Usage:

    >>> ctx = ProjectContext({
    ...     "foo": "bar"
    })
    >>> ctx["foo"]

    (from else where in the project)

    >>> from xyz.abc import jkl
    >>> assert jkl.ctx["foo"] == "bar"
    """

    def __hash__(self):
        return hash(tuple(sorted(self.items())))


ctx = ProjectContext()


def inactivated_iterator(genexp_iterator_factory=tuple):
    """
    A decorator, that ensures that the function
    returns another function, which when called,
    returns the actual result of the function.

    Done to ensure that a generator expression
    is not called before it is actually needed.

    Usage:

    >>> @inactivated_iterator(list)
    ... def foo(n):
    ...     yield from range(n)
    >>> foo(n=10)()
    ... [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    """

    def decorator(func):
        def wrapper(*args, **kwargs):
            def inactivated_function():
                retval = func(*args, **kwargs)
                if inspect.isgenerator(retval):
                    return genexp_iterator_factory(retval)
                return retval

            return functools.partial(inactivated_function)

        return wrapper

    return decorator


def iterate_modules_from_path(path: pathlib.Path, package: str, *, exempt: list = None):
    """
    Module iterator from a path for lazy loading.
    """
    if exempt is None:
        exempt = []

    for module_path in path.glob("*/"):
        if module_path.name not in exempt:
            yield module_path, importlib.import_module(
                "." + module_path.name, package=package
            )



"""
IntelliQ, a highly intelligent and robust quality string parser.

Supports:

- best
- 720/1080/2160
- best+subtitles
- titled+>=720[stream_url=r"abc"]
"""


import logging
from collections import namedtuple
from operator import eq, ge, gt, le, lt
from typing import Generator, Iterable, List, Optional, Tuple, TypeVar, Union

from animdl2.utils.optopt import regexlib

PARENTHESES = (
    ("(", ")"),
    ("[", "]"),
    ("{", "}"),
    ("«", "»"),
    ("‹", "›"),
    ("『", "』"),
    ("【", "】"),
    ("〖", "〗"),
    ("〘", "〙"),
    ("〚", "〛"),
    ("﴾", "﴿"),
    ("（", "）"),
    ("［", "］"),
    ("｛", "｝"),
    ("《", "》"),
    ("〈", "〉"),
    ("⦗", "⦘"),
    ("⦅", "⦆"),
    ("⦃", "⦄"),
)

QUOTES = ('"', "'")
ESCAPE = "\\"

PORTION_PARSER = regexlib.compile(
    r'(?P<key>.+?)=(?P<regex>r)?("|\')(?P<value>(?:\\\3|.)*?)\3'
)
SEGMENT_PARSER = regexlib.compile(r"(?i)(best|worst|\d+)?(.*)")

QUALITY_REGEX = regexlib.compile(
    r"(?i)(?:^|\b)(?P<expression>[<>=]=?)?(?:\d+x)?(?P<quality>\d+)p?\b"
)

EXPRESSION_MAPPING = {
    ">": gt,
    ">=": ge,
    "<": lt,
    "<=": le,
    "=": eq,
    "==": eq,
}

quality_operator = namedtuple(
    "quality_operator", ("quality", "operator"), defaults=(0, ge)
)


pair_value = TypeVar("pair_value")
stream_type = TypeVar("stream_type")


def remove_parentheses(value, n=None, *, parentheses=PARENTHESES):
    removed_count = 0
    has_parenthesis = True

    while has_parenthesis and (removed_count < n if n is not None else True):
        for start, end in parentheses:
            if value[0:1] == start and value[-1:] == end:
                value = value[1:-1]
                continue

        has_parenthesis = False

    return value


def evaluate_quality_to_number(quality: Optional[Union[str, int]]):
    if quality is None:
        return 0

    if isinstance(quality, int):
        return quality

    if quality.isdigit():
        return int(quality)

    match = QUALITY_REGEX.search(quality)

    if match is None:
        return 0

    return int(match.group("quality"))


def evaluate_quality_to_number_check(quality: Optional[Union[str, int]]):
    if quality is None:
        return None

    if isinstance(quality, int):
        return quality_operator(quality, le)

    if quality.isdigit():
        return quality_operator(int(quality), le)

    match = QUALITY_REGEX.search(quality)
    if match is None:
        return None

    return quality_operator(
        int(match.group("quality")),
        EXPRESSION_MAPPING.get(match.group("expression"), le),
    )


SPECIAL_SELECTORS = {
    "best": lambda streams: max(
        streams,
        key=lambda stream: evaluate_quality_to_number(stream.get("quality")),
    ),
    "worst": lambda streams: min(
        streams,
        key=lambda stream: evaluate_quality_to_number(stream.get("quality")),
    ),
}

SPECIAL_FILTERS = {
    "subtitles": lambda streams: [
        stream for stream in streams if stream.get("subtitle")
    ],
    "titled": lambda streams: [stream for stream in streams if stream.get("title")],
}


def iter_portions_by(
    string: str,
    splitters: Tuple[str] = ["/"],
    *,
    escape: str = ESCAPE,
    quoters: Tuple[str] = QUOTES,
    parenthesis: Tuple[Tuple[str, str]] = PARENTHESES,
):
    haystack = dict(parenthesis)
    haystack.update({_: _ for _ in quoters})
    escaping = False
    stack = []

    previous = None

    null_splitter = "" in splitters

    while string and string != previous:
        previous = string
        for n, content in enumerate(string):
            if not escaping:
                if content in haystack and stack:
                    if null_splitter:
                        yts = string[:n]

                        if yts:
                            yield yts

                        string = string[n:]

                    if stack and stack[-1] == content:
                        stack.pop()
                    else:
                        stack.append(content)

            escaping = content in escape

            if not stack and content in splitters:
                yield string[:n]
                string = string[n + 1 :]

    if string:
        yield string


def get_pair(
    target: "pair_value", pairs: Tuple[pair_value, pair_value]
) -> Tuple[Union[Tuple[pair_value, pair_value], Tuple[None, None], bool]]:
    for first, last in pairs:
        if target in (first, last):
            return (first, last), target == first

    return (None, None), False


def iter_fulfilling_streams(
    streams: Iterable[stream_type], quality_string: str, *, all=False
) -> Generator[stream_type, None, None]:
    streams = list(streams)

    for segment in iter_portions_by(quality_string):
        """
        Separating "best[subtitles]/best" into "best[subtitles]" and "best"
        """
        candidates = []

        for single_portion in iter_portions_by(segment, splitters=("",)):
            """
            Separating "best[subtitles]" into "best" and "[subtitles]"
            """
            for portion in iter_portions_by(single_portion, splitters=("+",)):
                for stream in candidates:
                    if stream in streams:
                        streams.remove(stream)

                case_insensitive_portion = portion.lower()

                quality_checker = evaluate_quality_to_number_check(
                    case_insensitive_portion
                )

                if quality_checker is not None:
                    for stream in streams:
                        if quality_checker.operator(
                            evaluate_quality_to_number(stream.get("quality")),
                            quality_checker.quality,
                        ):
                            candidates.append(stream)
                else:
                    candidates = streams[:]

                portion_match = PORTION_PARSER.search(remove_parentheses(portion))

                if portion_match is not None:
                    portion_key, regex, value = portion_match.group(
                        "key", "regex", "value"
                    )
                    if regex is not None:
                        compiled_regex = regexlib.compile(value)

                        candidates = [
                            stream
                            for stream in candidates
                            if portion_key in stream
                            and compiled_regex.search(stream[portion_key])
                        ]
                    else:
                        candidates = [
                            stream
                            for stream in candidates
                            if portion_key in stream and stream[portion_key] == value
                        ]

                if case_insensitive_portion in SPECIAL_FILTERS:
                    candidates = SPECIAL_FILTERS[case_insensitive_portion](candidates)

                if case_insensitive_portion in SPECIAL_SELECTORS:
                    candidates = [
                        SPECIAL_SELECTORS[case_insensitive_portion](candidates)
                    ]

                if candidates:
                    break

        for _ in candidates:
            yield single_portion, candidates

        if not all:
            return


def filter_quality(
    streams: Iterable[stream_type], quality_string: str
) -> List[stream_type]:
    portion_map = {}
    for portion, candidates in iter_fulfilling_streams(streams, quality_string):
        portion_map[portion] = candidates

    for portion, candidates in portion_map.items():
        if len(candidates):
            print(portion, candidates)
            return candidates

    return streams


if __name__ == "__main__":
    test_quality_strings = {
        "best": {
            "values": [
                {"quality": "2160p"},
                {"quality": "1080p"},
                {"quality": "720p"},
                {"quality": "480p"},
            ],
            "expected": [{"quality": "2160p"}],
        },
        "worst": {
            "values": [
                {"quality": "1080p"},
                {"quality": "720p"},
                {"quality": "480p"},
            ],
            "expected": [{"quality": "480p"}],
        },
        "=1920x1080p": {
            "values": [
                {"quality": "1080p"},
                {"quality": "720p"},
                {"quality": "480p"},
            ],
            "expected": [{"quality": "1080p"}],
        },
        "=1920x1080p+720p": {
            "values": [
                {"quality": "1080p"},
                {"quality": "720p"},
                {"quality": "480p"},
            ],
            "expected": [{"quality": "1080p"}],
        },
        "=720p": {
            "values": [
                {"quality": "1080p"},
                {"quality": "720p"},
                {"quality": "480p"},
            ],
            "expected": [{"quality": "720p"}],
        },
        "subtitles": {
            "values": [
                {"subtitle": [...]},
            ],
            "expected": [{"subtitle": [...]}],
        },
        "[stream_url=r'\.mp4$']": {
            "values": [
                {"stream_url": "abc.mp4"},
                {"stream_url": "abc.mpd"},
            ],
            "expected": [{"stream_url": "abc.mp4"}],
        },
    }

    for test, values in test_quality_strings.items():
        results = filter_quality(values["values"], test)

        assert results == values["expected"], f"{test!r} failed with {results}."

    print("All tests passed.")



"""
animdl: Utilities for the internal searching system. 
"""
import string
from typing import Callable, Iterable, Optional, TypeVar

from .optopt import regexlib

search_type = TypeVar("search_type")


options = {
    "ignore_whitespace": lambda x: regexlib.sub(r"\s+", "", x),
    "ignore_punctuation": lambda x: regexlib.sub(
        rf"[{regexlib.escape(string.punctuation)}]+", "", x
    ),
}


def iter_search_results(
    query: str,
    possibilities: Iterable[search_type],
    *,
    processor: Optional[Callable[[search_type], str]] = None,
    search_options: Iterable[str] = ("ignore_punctuation",),
):
    """
    Powerful searching function that uses regex building
    for matching the query with the possibilities.

    :param query: The query to search for.
    :param possibilities: The possibilities to search in.
    :param processor: A function that processes the possibilities.
    :param search_options: The options to use for searching.

    :return: A generator that yields the search results.
    """
    pattern = regexlib.compile(
        r"(.*?)".join(map(regexlib.escape, query.strip())),
        flags=regexlib.IGNORECASE,
    )

    def genexp():
        for search_value in possibilities:
            if processor is not None:
                processed_value = processor(search_value)
            else:
                processed_value: str = search_value  # type: ignore

            for option in search_options:
                if option in options:
                    search_value = options[option](search_value)

            match = pattern.search(processed_value)

            if match:
                yield len(processed_value), search_value

    for _, search_value in sorted(genexp(), key=lambda x: x[0]):
        yield search_value



import importlib.metadata

try:
    __core__ = importlib.metadata.version("animdl")
except importlib.metadata.PackageNotFoundError:
    import pathlib

    from animdl2.utils.optopt import regexlib

    __core__ = regexlib.search(
        r'name = "animdl"\nversion = "(.+?)"',
        (pathlib.Path(__file__).parent.parent / "pyproject.toml").read_text(),
    ).group(1)



"""
animdl: Utilities for the perfect HTTP disguise. 
"""

import sys
import typing as t
from urllib.parse import quote, urlencode

from animdl2.core.__version__ import __core__

if t.TYPE_CHECKING:
    import logging

    import httpx

CORS_PROXY = "https://corsproxy.io/"


def get_user_agent():
    """
    Returns the user agent to be used in the requests.

    The project will not be detected as a bot so
    standardising is better for the web developers.

    :return: The user agent.
    """
    return f"animdl/{__core__} ({sys.platform})"


def cors_proxify(
    session: "httpx.Client",
    method: str,
    url: str,
    *args,
    params=None,
    headers=None,
    **kwargs,
):
    """
    CORS Proxy is a service that forwards the requests
    to the target website and returns the response.

    This may also bypass geo-restrictions.

    :param session: The session to use for the request.
    :param method: The method to use for the request.
    :param url: The URL to send the request to.
    :param args: The arguments to pass to the request.
    :param params: The parameters to pass to the request.
    :param headers: The headers to pass to the request.
    :param kwargs: The keyword arguments to pass to the request.

    :return: The response from the request.
    """
    if params is not None:
        url += "?" + urlencode(params)

    cors_url = CORS_PROXY + "?" + quote(url)

    if headers is None:
        headers = {}

    headers.update(referer=url)
    return session.request(method, cors_url, headers=headers, *args, **kwargs)


def integrate_ddg_bypassing(session: "httpx.Client", *hosts: str):
    """
    DDoS Guard is a service that protects websites from
    DDoS attacks. It does this by blocking the requests
    from the bots and crawlers. This function integrates
    the bypassing of the DDoS Guard into the session.

    :param session: The session to integrate the bypassing into.
    :param hosts: The hosts to bypass the DDoS Guard for.

    :return: None
    """
    for host in hosts:
        session.cookies.set(
            "__ddg2_",
            "YW5pbWRsX3NheXNfaGkNCg.",
            domain=host,
        )


def setup_global_http_exception_hook(
    exit_code: int,
    http_error_baseclass: t.Type["httpx.HTTPError"],
    logger: t.Optional["logging.Logger"] = None,
    msg: str = (
        "{value!r}, this issue originates due to connection issues on your or the server's side. "
        "Retry after troubleshooting connection issues on your system."
    ),
):
    """
    Sets up the global exception hook for any HTTP exceptions
    for graceful exit.

    :return: None
    """
    hook = sys.excepthook

    def exception_hook(exctype, value, traceback):
        if issubclass(exctype, http_error_baseclass):
            if logger is not None:
                logger.error(
                    msg.format(value=value),
                    exc_info=True,
                )

            exit(exit_code)

        return hook(exctype, value, traceback)

    sys.excepthook = exception_hook


if __name__ == "__main__":
    import httpx

    client = httpx.Client(
        headers={
            "User-Agent": get_user_agent(),
        },
        timeout=30.0,
        follow_redirects=True,
    )

    integrate_ddg_bypassing(
        client,
        ".marin.moe",
        ".haho.moe",
    )

    assert (
        client.get("https://marin.moe/").status_code == 200
    ), "Could not bypass DDoS Guard."
    assert (
        client.get(
            "http://www.crunchyroll.com/",
            headers={
                "Referer": "",
            },
        ).status_code
        == 200
    ), "Could not bypass cloudflare."
    assert (
        cors_proxify(
            client,
            "GET",
            "https://www.animeout.xyz/",
        ).status_code
        == 200
    ), "Could not validate CORS Proxy."

    print("HTTP Client is primed.", file=sys.stderr)



"""
animdl: Utility functions for caching HTTP requests.
"""

import atexit
import os
import pathlib
import sqlite3
import time
from datetime import timedelta
from urllib.parse import urlencode

import httpx

from .optopt import jsonlib, regexlib

caching_mimetypes = {
    "application": {
        "json",
        "xml",
        "x-www-form-urlencoded",
        "x-javascript",
        "javascript",
    },
    "text": {"html", "css", "javascript", "plain", "xml", "xsl", "x-javascript"},
}


class CachingHTTPXClient(httpx.Client):
    """
    HTTPX client with caching capabilities.

    :param cache_db_path: The path to the cache database.
    :param cache_db_lock_file: The path to the cache database lock file.
    :param max_lifetime: The maximum lifetime of a cache entry.
    :param max_size: The maximum size of the cache database.
    :param table_name: The name of the table to use for the cache.

    This class adds a middleware to the HTTPX client that
    caches the responses to an sqlite3 database.

    The cache database is a single table with the following
    columns:

    - url: The URL of the request.
    - status_code: The status code of the response.
    - request_headers: The request headers of the request.
    - response_headers: The response headers of the response.
    - data: The data of the response.
    - redirection_policy: The redirection policy of the request.
    - cache_expiry: The expiry time of the cache entry.

    Usage:

    >>> from animdl2.utils.http_caching import CachingHTTPXClient
    >>> import time
    >>> client = CachingHTTPXClient("cache.db", "cache.lock")
    >>> initial_time = time.time()
    >>> client.get("https://example.com")
    <Response [200 OK]>
    >>> print("Request took:", time.time() - initial_time)
    >>> initial_time = time.time()
    >>> client.get("https://example.com")
    <Response [200 OK]>
    >>> print("Cached request took:", time.time() - initial_time)
    """

    request_functions = (
        "get",
        "options",
        "head",
        "post",
        "put",
        "patch",
        "delete",
    )

    def __new__(cls, *args, **kwargs):
        def caching_params(name: str):
            def wrapper(self, *args, **kwargs):
                return cls.request(self, name.upper(), *args, **kwargs)

            return wrapper

        for func in cls.request_functions:
            setattr(cls, func, caching_params(func))

        return super().__new__(cls)

    def __init__(
        self,
        cache_db_path: str,
        cache_db_lock_file: str,
        max_lifetime: int = 604800,
        max_size: int = (1024**2) * 10,
        table_name: str = "animdl_http_cache",
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.cache_db_path = cache_db_path
        self.max_lifetime = max_lifetime
        self.max_size = max_size
        self.connection = sqlite3.connect(self.cache_db_path)
        self.table_name = table_name

        self.connection.cursor().execute(
            f"create table if not exists {self.table_name!r} ("
            "url text, "
            "status_code integer, "
            "request_headers text, "
            "response_headers text, "
            "data blob,"
            "redirection_policy int,"
            "cache_expiry integer"
            ")",
        )

        atexit.register(lambda conn: conn.commit() or conn.close(), self.connection)
        self.setup_lockfile(cache_db_lock_file)

    def request(
        self,
        method,
        url,
        params=None,
        force_caching=False,
        fresh=False,
        *args,
        **kwargs,
    ):

        if fresh:
            return super().request(method, url, params=params, *args, **kwargs)

        if params is not None:
            url += "?" + urlencode(params)

        redirection_policy = int(kwargs.get("force_redirects", False))
        cursor = self.connection.cursor()

        cursor.execute(
            "select "
            "status_code, "
            "request_headers, "
            "response_headers, "
            "data, "
            "redirection_policy "
            f" from {self.table_name!r} where url = ? and redirection_policy = ? and cache_expiry > ?",
            (url, redirection_policy, int(time.time())),
        )

        result = cursor.fetchone()

        if result is not None:
            (
                status_code,
                request_headers,
                response_headers,
                data,
                redirection_policy,
            ) = result

            response = httpx.Response(
                status_code=status_code,
                headers=jsonlib.loads(response_headers),
                content=data,
                request=httpx.Request(
                    method, url, headers=jsonlib.loads(request_headers), *args, **kwargs
                ),
            )
            response.elapsed = timedelta(seconds=0)
            return response

        response = super().request(method, url, *args, **kwargs)

        if response.status_code < 400 and (
            force_caching
            or self.is_content_type_cachable(
                response.headers.get("content-type"), caching_mimetypes
            )
            and len(response.content) < self.max_size
        ):

            cursor.execute(
                f"insert into {self.table_name!r} values (?, ?, ?, ?, ?, ?, ?)",
                (
                    url,
                    response.status_code,
                    jsonlib.dumps(dict(response.request.headers)),
                    jsonlib.dumps(dict(response.headers)),
                    response.content,
                    redirection_policy,
                    int(time.time()) + self.max_lifetime,
                ),
            )

            self.connection.commit()

        return response

    @staticmethod
    def is_content_type_cachable(content_type, caching_mimetypes):
        if content_type is None:
            return True

        mime, contents = content_type.split("/")

        contents = regexlib.sub(r";.*$", "", contents)

        return mime in caching_mimetypes and any(
            content in caching_mimetypes[mime] for content in contents.split("+")
        )

    @staticmethod
    def setup_lockfile(lock_file: str):
        lockfile_path = pathlib.Path(lock_file)

        if lockfile_path.exists():
            with lockfile_path.open("r") as f:
                pid = f.read()

            raise RuntimeError(
                f"This instance of {__class__.__name__!r} is already running with PID: {pid}. "
                "Sqlite3 does not support multiple connections to the same database. "
                "If you are sure that no other instance of this class is running, "
                f"delete the lock file at {lockfile_path.as_posix()!r} and try again."
            )

        with lockfile_path.open("w") as f:
            f.write(str(os.getpid()))

        atexit.register(lockfile_path.unlink)


if __name__ == "__main__":
    client = CachingHTTPXClient("cache.db", "cache.lockfile")

    print(
        client.get(
            "https://google.com",
        ).elapsed
    )

    print(
        client.get(
            "https://google.com",
        ).elapsed
    )



from collections import defaultdict
from functools import partial

import yarl

from ....config import ALLANIME, SUPERANIME_RETURN_ALL, SUPERANIME_TYPE_OF
from ...helpers import construct_site_based_regex, optopt, superscrapers
from .gql_api import api

REGEX = construct_site_based_regex(ALLANIME, extra_regex=r"/(?:anime|watch)/([^?&/]+)")

LEGACY_CONTENT_REGEX = optopt.regexlib.compile(
    r'date:g,data:"({.+?})"}},calling:', optopt.regexlib.DOTALL
)


def iter_episodes(
    episode_dictionary: dict,
    *,
    selected_type_of: str = SUPERANIME_TYPE_OF,
):
    episodes = defaultdict(list)

    for type_of, episode_numbers in episode_dictionary.items():
        if type_of != selected_type_of:
            continue

        for episode in episode_numbers:
            episodes[int(episode) if episode.isdigit() else 0].append(
                (type_of, episode)
            )

    yield from sorted(episodes.items(), key=lambda x: x[0])


def to_clock_json(url: str):
    parsed = yarl.URL(url)
    return parsed.with_name("clock.json").with_query(parsed.query).human_repr()


def fetch_legacy(session, episode, type_of, show_id):
    """
    This is a fallback method for fetching the content of an episode
    when the GQL hits server-side rate-limits.
    """

    match = LEGACY_CONTENT_REGEX.search(
        session.get(
            ALLANIME + f"watch/{show_id}//episode-{episode}-{type_of}",
        ).text
    )

    if match:
        return (
            optopt.jsonlib.loads(match.group(1).replace('\\"', '"'))
            .get("data", {})
            .get("episode", {})
        )

    return {}


def one_digit_symmetric_xor(password: int, target: str):
    def genexp():
        for segment in bytearray.fromhex(target):
            yield segment ^ password

    return bytes(genexp()).decode("utf-8")


def extract_content(
    session,
    content: "iter_episodes",
    show_id: str,
    *,
    api_endpoint: str,
    return_all: bool = SUPERANIME_RETURN_ALL,
):
    for type_of, episode in content:
        api_response = api.fetch_episode(
            session, show_id, episode, translation_type=type_of
        )

        episode_info = api_response.get("episode") or fetch_legacy(
            session, episode, type_of, show_id
        )

        if not episode_info:
            continue

        attrs = {}

        if "notes" in episode_info and episode_info["notes"]:
            attrs.update(title=episode_info["notes"].replace("<note-split>", " // "))

        sources = sorted(
            episode_info["sourceUrls"],
            key=lambda value: value.get("priority", 0),
            reverse=True,
        )

        if not sources:
            continue

        for source in sources:
            if source["sourceUrl"][:2] == "--":
                source["sourceUrl"] = one_digit_symmetric_xor(
                    56, source["sourceUrl"][2:]
                )

            if source["sourceUrl"][0] == "/":
                response = session.get(
                    to_clock_json(api_endpoint + source["sourceUrl"])
                )

                if response.text in ("error", "error no result"):
                    continue

                streams = response.json().get("links")

                if not streams:
                    continue

                for stream_data in streams:
                    subtitles = list(
                        map(
                            lambda subtitles: subtitles["src"],
                            stream_data.get("subtitles", []),
                        )
                    )

                    attrs.setdefault("subtitle", [])
                    attrs["subtitle"].extend(subtitles)

                    if "rawUrls" in stream_data and stream_data["rawUrls"]:
                        audio_tracks = list(
                            map(
                                lambda data: data["url"],
                                stream_data["rawUrls"].get("audios", []),
                            )
                        )

                        for data in stream_data["rawUrls"]["vids"]:
                            yield {
                                "quality": data["height"],
                                "stream_url": data["url"],
                                "audio_tracks": audio_tracks,
                                **attrs,
                            }

                        continue

                    if "link" in stream_data and stream_data["link"]:
                        url = stream_data["link"]
                    else:
                        if not stream_data["portData"]["streams"]:
                            continue

                        stream = stream_data["portData"]["streams"][0]

                        url = stream["url"]

                    yield from superscrapers.iter_unpacked_from_packed_hls(
                        session, yarl.URL(url), stream_attribs=attrs
                    )

                    if not return_all:
                        return

            else:
                yield from superscrapers.iter_unpacked_from_packed_hls(
                    session, yarl.URL(source["sourceUrl"]), stream_attribs=attrs
                )
                if not return_all:
                    return


def fetcher(session, url: "str", check, match):
    api_endpoint = (
        session.get(ALLANIME + "getVersion").json().get("episodeIframeHead", "")
    )

    available_episodes = api.fetch_show_info(
        session,
        match.group(1),
    ).get("availableEpisodesDetail", {})

    for episode, content in iter_episodes(available_episodes):
        if check(episode):
            yield partial(
                lambda session, content, show_id: list(
                    extract_content(
                        session, content, show_id, api_endpoint=api_endpoint
                    )
                ),
                session,
                content,
                match.group(1),
            ), episode


def metadata_fetcher(session, url: "str", match):
    anime_name = api.fetch_show_info(
        session,
        match.group(1),
    ).get("name", "")

    return {"titles": [anime_name]}



import base64
from urllib.parse import quote, unquote

from Cryptodome.Cipher import ARC4

VRF_KEY = b"iECwVsmW38Qe94KN"
URL_KEY = b"hlPeNwkncH0fq9so"
CHAR_SUBST_OFFSETS = (-4, -4, 3, 3, 6, -4, 3, -6, -2, -4)


VRF_ENDPOINT = "https://9anime.eltik.net"
VRF_API_KEY = "jerry"  # Remind me to get my own key someday


def vrf_api(session, content_id: str, api_key=VRF_API_KEY, *, endpoint="vrf"):
    return session.get(
        VRF_ENDPOINT + f"/{endpoint}",
        params={"query": content_id, "apikey": api_key},
    ).json()["url"]


def vrf_futoken(session, futoken, endpoint, query):
    return session.post(
        VRF_ENDPOINT + f"/{endpoint}",
        params={
            "query": query,
            "apikey": VRF_API_KEY,
        },
        data={
            "futoken": futoken,
            "query": query,
        },
    ).json()["rawURL"]


def char_subst(content: bytes, *, offsets=CHAR_SUBST_OFFSETS):
    for n, value in enumerate(content):
        yield (value + offsets[n % len(offsets)])


def generate_vrf_from_content_id(
    content_id: str,
    key=VRF_KEY,
    *,
    offsets=CHAR_SUBST_OFFSETS,
    encoding="utf-8",
    reverse_later=True,
):
    encoded_id = base64.b64encode(
        ARC4.new(key).encrypt(quote(content_id).encode(encoding))
    )

    if reverse_later:
        encoded_id = bytes(char_subst(encoded_id, offsets=offsets))[::-1]
    else:
        encoded_id = bytes(char_subst(encoded_id[::-1], offsets=offsets))

    return base64.b64encode(encoded_id).decode(encoding)


def decrypt_url(encrypted_url: str, key=URL_KEY, *, encoding="utf-8"):
    return unquote(
        ARC4.new(key).decrypt(base64.b64decode(encrypted_url)).decode(encoding)
    )



from functools import partial

import lxml.html as htmlparser
import regex

from ....config import HENTAISTREAM
from ...helpers import construct_site_based_regex

EPISODE_REGEX = regex.compile(r"/\d+/[^&?/]+")
TITLES_REGEX = regex.compile(r'<h1 class="entry-title" itemprop="name">(.+?)</h1>')


REGEX = construct_site_based_regex(HENTAISTREAM, extra_regex=r"/hentai/([^?&/]+)")

EXTRACTOR_REGEX = regex.compile(r"src: '(.+?)',\s+type: '.+?',\s+size: (\d+),")
SUBTITLE_URL_REGEX = regex.compile(r"subUrl: '(.+?)'")


def fetch_streams(session, episode_url, **opts):
    episode_page = session.get(HENTAISTREAM + episode_url[1:]).text
    subtitles = SUBTITLE_URL_REGEX.findall(episode_page)

    for url, size in EXTRACTOR_REGEX.findall(episode_page):
        yield {
            "stream_url": url,
            "quality": int(size),
            "headers": {"referer": HENTAISTREAM},
            "subtitle": subtitles,
        }


def fetcher(session, url, check, match):
    response = session.get(match.group(0))
    episode_page = htmlparser.fromstring(response.text)

    episode_id = episode_page.cssselect("input#e_id")[0].get("value")
    csrf = response.cookies.get("XSRF-TOKEN")

    data = session.post(
        HENTAISTREAM + "player/api",
        json={
            "episode_id": episode_id,
        },
        headers={
            "x-xsrf-token": csrf.replace("%3D", "="),
        },
    ).json()

    resolutions = (1080, 720)

    if data["resolution"] == "4k":
        resolutions = (2160,) + resolutions

    yield partial(
        lambda _: _,
        [
            {
                "title": data["title"],
                "stream_url": data["stream_domains"][-1]
                + "/"
                + data["stream_url"]
                + f"/{resolution}/manifest.mpd",
                "subtitle": [
                    data["stream_domains"][-1] + "/" + data["stream_url"] + f"/eng.ass",
                ],
                "quality": resolution,
            }
            for resolution in resolutions
        ],
    ), 1


def metadata_fetcher(session, url, match):
    page = htmlparser.fromstring(session.get(match.group(0)).text)
    return {"titles": page.cssselect("h1.leading-tight")[0].text_content().strip()}



import logging

import httpx

from animdl2 import utils

from .exit_codes import INTERNET_ISSUE

http_logger = logging.getLogger("http-client")

client = httpx.Client(
    headers={"user-agent": "animdl/1.0.0"},
    timeout=30,
    follow_redirects=True,
)

utils.http_client.integrate_ddg_bypassing(
    client,
    ".marin.moe",
)

utils.http_client.setup_global_http_exception_hook(
    exit_code=INTERNET_ISSUE,
    http_error_baseclass=httpx.HTTPError,
    logger=http_logger,
)

setattr(client, "cf_request", utils.http_client.cors_proxify)



import os
import sys
import warnings
from pathlib import Path

import yaml


def merge_dicts(dict1, dict2):
    for k, v in dict1.items():
        if isinstance(v, dict):
            merge_dicts(v, dict2.setdefault(k, {}))
        else:
            if k not in dict2:
                dict2[k] = v
    return dict2


def get_existent_path(*user_paths):
    for _ in user_paths:
        if not isinstance(_, Path):
            _ = Path(_)
        if _.exists():
            return _


if sys.platform == "win32":
    USERPROFILE_ANIMDL_PATH = (
        Path(os.getenv("LOCALAPPDATA", ".")) / ".config" / "animdl" / "config.yml"
    )
    OLD_DEPRECATED_PATH = Path(os.getenv("USERPROFILE")) / ".animdl" / "config.yml"

    if OLD_DEPRECATED_PATH.exists():
        if not USERPROFILE_ANIMDL_PATH.exists():
            warnings.warn(
                f"The config file path @ {OLD_DEPRECATED_PATH.as_posix()} is deprecated and will be removed in the future. "
                f"Please migrate to {USERPROFILE_ANIMDL_PATH.as_posix()}. "
                f"This is not done automatically because this project does not want to mess with your files.",
            )
            USERPROFILE_ANIMDL_PATH = OLD_DEPRECATED_PATH

else:
    USERPROFILE_ANIMDL_PATH = (
        Path(os.getenv("HOME", ".")) / ".config" / "animdl" / "config.yml"
    )

CONFIGURATION_FILE_PATH = get_existent_path(
    os.getenv("ANIMDL_CONFIG", "./animdl_config.yml"),
    "/animdl_config.yml",
    USERPROFILE_ANIMDL_PATH,
)

DEFAULT_CONFIG = {
    "default_provider": "allanime",
    "site_urls": {
        "9anime": "https://aniwave.to/",
        "allanime": "https://allanime.to/",
        "animekaizoku": "https://animekaizoku.com/",
        "animeout": "https://www.animeout.xyz/",
        "animepahe": "https://animepahe.ru/",
        "animeonsen": "https://animeonsen.xyz/",
        "animexin": "https://animexin.xyz/",
        "animixplay": "https://animixplay.to/",
        "animtime": "https://animtime.com/",
        "crunchyroll": "http://www.crunchyroll.com/",
        "kawaiifu": "https://kawaiifu.com/",
        "gogoanime": "https://gogoanime.bid/",
        "haho": "https://haho.moe/",
        "hentaistream": "https://hstream.moe/",
        "kamyroll_api": "https://kamyroll.herokuapp.com/",
        "marin": "https://marin.moe/",
        "twist": "https://twist.moe/",
        "yugen": "https://yugenanime.ro/",
        "zoro": "https://aniwatch.to/",
    },
    "quality_string": "best[subtitle]/best",
    "default_player": "mpv",
    "players": {
        "mpv": {
            "executable": "mpv",
            "opts": [],
        },
        "vlc": {
            "executable": "vlc",
            "opts": [],
        },
        "iina": {
            "executable": "iina",
            "opts": ["--keep-running"],
        },
        "celluloid": {
            "executable": "celluloid",
            "opts": [],
        },
        "ffplay": {
            "executable": "ffplay",
            "opts": [],
        },
        "android": {
            "executable": "am",
            "opts": [],
        },
    },
    "schedule": {
        "site_url": "https://graphql.anilist.co/",
        "date_format": "%b. %d, %A",
        "time_format": "%X",
    },
    "download_auto_retry": 300,
    "ffmpeg": {
        "executable": "ffmpeg",
        "use_for_downloads": False,
        "submerge": True,
    },
    "discord_presence": False,
    "fzf": {
        "executable": "fzf",
        "opts": [],
        "state": False,
    },
    "download_directory": ".",
    "threaded_download": False,
    "check_for_updates": True,
    "force_streaming_quality_selection": True,
    "aniskip": False,
    "superanime": {
        "return_all": False,
        "type_of": "sub",
        "crunchyroll": {"subtitle_language": "en-US"},
    },
}

CONFIG = DEFAULT_CONFIG

if CONFIGURATION_FILE_PATH is not None:
    with open(CONFIGURATION_FILE_PATH, "r") as conf:
        CONFIG = merge_dicts(DEFAULT_CONFIG, yaml.load(conf, Loader=yaml.SafeLoader))

SITE_URLS = CONFIG.get("site_urls", {})

NINEANIME = SITE_URLS.get("9anime")
ALLANIME = SITE_URLS.get("allanime")
ANIMEKAIZOKU = SITE_URLS.get("animekaizoku")
ANIMEOUT = SITE_URLS.get("animeout")
ANIMEPAHE = SITE_URLS.get("animepahe")
ANIMEONSEN = SITE_URLS.get("animeonsen")
ANIMEXIN = SITE_URLS.get("animexin")
ANIMIXPLAY = SITE_URLS.get("animixplay")
ANIMTIME = SITE_URLS.get("animtime")
CRUNCHYROLL = SITE_URLS.get("crunchyroll")
KAWAIIFU = SITE_URLS.get("kawaiifu")
KAMYROLL_API = SITE_URLS.get("kamyroll_api")
GOGOANIME = SITE_URLS.get("gogoanime")
MARIN = SITE_URLS.get("marin")
HAHO = SITE_URLS.get("haho")
HENTAISTREAM = SITE_URLS.get("hentaistream")
TWIST = SITE_URLS.get("twist")
YUGEN = SITE_URLS.get("yugen")
ZORO = SITE_URLS.get("zoro")

QUALITY = CONFIG.get("quality_string")

DEFAULT_PLAYER = CONFIG.get("default_player")
PLAYERS = CONFIG.get("players")

ANICHART = CONFIG.get("schedule", {}).get("site_url")

DATE_FORMAT = CONFIG.get("schedule", {}).get("date_format")
TIME_FORMAT = CONFIG.get("schedule", {}).get("time_format")

DEFAULT_PROVIDER = CONFIG.get("default_provider")

AUTO_RETRY = CONFIG.get("download_auto_retry", 300) / 1000

FFMPEG_SETTINGS = CONFIG.get("ffmpeg", {})

FFMPEG_EXECUTABLE = FFMPEG_SETTINGS.get("executable", "ffmpeg")
FFMPEG_USE_FOR_DOWNLOADS = FFMPEG_SETTINGS.get("use_for_downloads", False)
FFMPEG_SUBMERGE = FFMPEG_SETTINGS.get("submerge", True)

DISCORD_PRESENCE = CONFIG.get("discord_presence", False)

FZF = CONFIG.get("fzf", {})
FZF_EXECUTABLE = FZF.get("executable", "fzf")
FZF_OPTS = FZF.get("opts", [])
FZF_STATE = FZF.get("state", False)

DOWNLOAD_DIRECTORY = CONFIG.get("download_directory", ".")
THREADED_DOWNLOAD = CONFIG.get("threaded_download", False)

CHECK_FOR_UPDATES = CONFIG.get("check_for_updates", True)
FORCE_STREAMING_QUALITY_SELECTION = CONFIG.get(
    "force_streaming_quality_selection", True
)

USE_ANISKIP = CONFIG.get("aniskip", False)

SUPERANIME = CONFIG.get("superanime", {})

SUPERANIME_RETURN_ALL = SUPERANIME.get("return_all", False)
SUPERANIME_TYPE_OF = SUPERANIME.get("type_of", "sub")
SUPERANIME_CRUNCHYROLL = SUPERANIME.get("crunchyroll", {"subtitles": "en-US"})



"""
Internet Download Manager Support.

This is made strictly for Windows.
"""

import pathlib
import time
import typing
from collections import namedtuple

import rich

authentication = namedtuple(
    "authentication", "username password", defaults=(None, None)
)


idmlib, idmlib_object = None, None


try:
    from comtypes import client

    idmlib = client.GetModule(["{ECF21EAB-3AA8-4355-82BE-F777990001DD}", 1, 0])

    idmlib_object = client.CreateObject(
        idmlib.CIDMLinkTransmitter, None, None, idmlib.ICIDMLinkTransmitter2
    )
except Exception:
    pass


def supported():
    return bool(idmlib)


def idm_download(
    url: str,
    *,
    headers: typing.Optional[typing.Dict[str, str]] = None,
    form_data: typing.Optional[str] = None,
    auth: authentication = authentication(),
    filename: typing.Optional[str] = None,
    download_folder: pathlib.Path = pathlib.Path("."),
    lflag: int = 5,
):
    return idmlib_object.SendLinkToIDM(
        url,
        headers.get("referer"),
        headers.get("cookie"),
        form_data,
        auth.username,
        auth.password,
        download_folder.resolve(strict=True).as_posix(),
        filename,
        lflag,
    )


def wait_until_download(
    url,
    download_folder: pathlib.Path,
    headers=None,
    form_data=None,
    auth=authentication(),
    filename=None,
):
    idm_download(
        url,
        headers=headers,
        form_data=form_data,
        auth=auth,
        filename=filename,
        download_folder=download_folder,
        lflag=5,
    )

    if filename is None:
        return

    while not (download_folder / filename).exists():
        try:
            time.sleep(0.5)
        except KeyboardInterrupt:
            console = rich.get_console()

            console.print(
                "[bold red]Detected KeyboardInterrupt[/], continue or terminate?",
            )

            if console.input("[C/t]: ").lower() == "t":
                raise

            return



import logging
import os
import pathlib
import tempfile
import typing

import httpx
from tqdm import tqdm

from animdl2.utils import media_downloader, serverfiles

from ...config import (
    FFMPEG_EXECUTABLE,
    FFMPEG_SUBMERGE,
    FFMPEG_USE_FOR_DOWNLOADS,
    THREADED_DOWNLOAD,
)
from .ffmpeg import FFMPEG_EXTENSIONS, ffmpeg_download, has_ffmpeg, merge_subtitles
from .hls import HLS_STREAM_EXTENSIONS, hls_yield

if FFMPEG_EXECUTABLE:
    ffmpeg_executable = FFMPEG_EXECUTABLE


DEFAULT_MEDIA_EXTENSION = "mp4"
POSSIBLE_VIDEO_EXTENSIONS = (
    "mp4",
    "mkv",
    "webm",
    "flv",
    "avi",
    "mov",
    "wmv",
    "mpg",
    "mpeg",
    "m4v",
    "3gp",
    "3g2",
    "m2v",
    "m4v",
    "f4v",
    "f4p",
    "f4a",
    "ts",
    "m3u8",
    "m3u",
    "mpd",
    "ism",
    "ismv",
    "av1",
)


def sanitize_filename(f):
    return "".join(" - " if _ in '<>:"/\\|?*' else _ for _ in f).strip()


def standard_download(
    session: httpx.Client,
    url: str,
    expected_download_path: pathlib.Path,
    content_size: "typing.Optional[int]" = None,
    headers: "typing.Optional[dict]" = None,
    ranges: bool = True,
    log_level: int = 20,
    retry_timeout: float = 5.0,
    method="GET",
):
    expected_download_path.parent.mkdir(parents=True, exist_ok=True)

    logger = logging.getLogger(f"downloader/standard[{expected_download_path.name}]")

    with open(expected_download_path, "ab") as download_io:
        position = download_io.tell()

        if position:
            if not ranges:
                logger.critical(
                    "Stream does not support ranged downloading, the previous download cannot be continued."
                )
                download_io.seek(0)
                position = 0

            if THREADED_DOWNLOAD:
                logger.critical(
                    "Download will not be continued as continuation is not supported in threaded downloads."
                )
                download_io.seek(0)
                position = 0

        progress_bar = tqdm(
            desc=f"{method} / {expected_download_path.name}",
            total=content_size,
            disable=log_level > 20,
            initial=position,
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
        )

        downloader = media_downloader.MultiConnectionDownloader(
            session,
            method,
            url,
            headers=headers,
            retry_timeout=retry_timeout,
            progress_bar=progress_bar,
        )

        downloader.allocate_downloads(
            download_io,
            content_size,
            start_at=position,
            threaded=ranges and THREADED_DOWNLOAD,
        )

    progress_bar.close()
    return expected_download_path


def hls_download(
    session: httpx.Client,
    url: str,
    expected_download_path: pathlib.Path,
    headers: dict = {},
    method="GET",
    **opts,
):
    content_path = expected_download_path.with_suffix(".mpeg")
    index_holder = expected_download_path.with_suffix(".index")
    index = 1

    if index_holder.exists():
        with open(index_holder, "r") as ih:
            index = int(ih.read() or 1)

    sizes = []

    with open(content_path, "ab") as tsstream, open(index_holder, "w+") as istream:
        downloaded = tsstream.tell()
        if downloaded:
            sizes.extend([downloaded / index] * index)
        progress_bar = tqdm(
            desc=f"HLS {method} / {content_path.name}",
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
            initial=downloaded,
            disable=opts.get("log_level", 20) > 20,
        )

        for content in hls_yield(
            session,
            [{"stream_url": url, "headers": headers}],
            opts.get("preferred_quality") or 1080,
            opts.get("retry_timeout") or 5,
            continuation_index=index,
        ):
            stream = content.get("bytes")
            total = content.get("total")
            current = content.get("current")

            size = len(stream)
            sizes.append(size)

            mean = sum(sizes) / len(sizes)
            stddev = (sum(abs(mean - s) ** 2 for s in sizes) / len(sizes)) ** 0.5

            progress_bar.total = mean * total
            progress_bar.desc = "HLS GET / {}.ts [± {:.3f} MB]".format(
                content_path, stddev / (1024**2)
            )
            progress_bar.update(size)

            istream.write(str(current or 0 + 1))
            istream.seek(0)
            istream.flush()

            tsstream.write(stream)

    progress_bar.close()
    os.remove(index_holder)
    return content_path


def idm_download(url, headers, expected_download_path: pathlib.Path, **opts):
    from .idmanlib import wait_until_download as idmdl

    idmdl(
        url,
        headers=headers or {},
        download_folder=expected_download_path.parent,
        filename=expected_download_path.name,
    )
    return expected_download_path


def subautomatic(f: "handle_download"):
    def __inner__(
        session: httpx.Client,
        url: str,
        expected_download_path: pathlib.Path,
        use_internet_download_manager=False,
        headers: typing.Optional[typing.Dict[str, str]] = None,
        **opts,
    ):
        logger = logging.getLogger("ffmpeg/submerge")

        subtitles = opts.pop("subtitles", [])

        content_path = f(
            session=session,
            url=url,
            expected_download_path=expected_download_path,
            use_internet_download_manager=use_internet_download_manager,
            headers=headers,
            **opts,
        )  # type: pathlib.Path

        if not subtitles:
            return content_path

        locations = []

        for n, subtitle in enumerate(subtitles):
            prefetch_response = media_downloader.prefetch(
                session, opts.get("method", "GET"), subtitle, headers=headers
            )

            content_disposition = prefetch_response.headers.get("Content-Disposition")

            if content_disposition:
                server_filename = serverfiles.guess_from_content_disposition(
                    content_disposition
                )
            else:
                content_type = prefetch_response.headers.get("Content-Type")

                if content_type and content_type != "application/octet-stream":
                    server_filename = serverfiles.guess_from_content_type(
                        None, content_type
                    )

                else:
                    server_filename = serverfiles.guess_from_path(
                        prefetch_response.url.path
                    )

            download_location = content_path.parent / (
                f"{content_path.name}_sub_{n}_" + (server_filename or "")
            )

            standard_download(
                session,
                subtitle,
                download_location,
                headers=headers,
                log_level=opts.get("log_level", 20),
                retry_timeout=opts.get("retry_timeout", 5),
            )

            locations.append(download_location.resolve().as_posix())

        if (not has_ffmpeg()) or not FFMPEG_SUBMERGE:
            logger.info(
                f"ffmpeg is unavailable, subtitle is downloaded to {download_location.as_posix()}."
            )
            return content_path

        out_path = content_path.with_name(
            content_path.name + "_submerged" + content_path.suffix
        )

        ffmpeg_returncode = merge_subtitles(
            content_path,
            out_path,
            locations,
            log_level=opts.get("log_level", 20),
        )

        if ffmpeg_returncode:
            logger.warning(
                f"Got a non-zero return code {ffmpeg_returncode} from ffmpeg, the original file will not be erased in case of merge failure."
            )
            return content_path

        content_path.unlink(missing_ok=True)

        try:
            out_path.rename(content_path)
        except OSError:
            logger.warning(
                f"Failed to rename {out_path} to {content_path}, the converted file will not be erased in case of merge failure."
            )

        return out_path

    return __inner__


@subautomatic
def handle_download(
    session: httpx.Client,
    url: str,
    expected_download_path: pathlib.Path,
    use_internet_download_manager=False,
    headers: typing.Optional[typing.Dict[str, str]] = None,
    **opts,
):
    download_handling_logger = logging.getLogger("animdl/download-handler")

    prefetch_response = media_downloader.prefetch(
        session, opts.get("method", "GET"), url, headers=headers
    )

    content_disposition = prefetch_response.headers.get("Content-Disposition")

    server_filename = None

    if server_filename is None and content_disposition:
        server_filename = serverfiles.guess_from_content_disposition(
            content_disposition
        )

    content_type = prefetch_response.headers.get("Content-Type")

    if server_filename is None and content_type:
        server_filename = serverfiles.guess_from_content_type(None, content_type)

    if server_filename is None:
        server_filename = serverfiles.guess_from_path(prefetch_response.url.path)

    content_range = prefetch_response.headers.get("Content-Range")

    if content_range is not None:
        content_size = int(content_range.split("/", 1)[1])
    else:
        content_size = None

    ranges = prefetch_response.status_code == 206 or (
        prefetch_response.headers.get("Accept-Ranges") == "bytes"
    )

    if server_filename and "." in server_filename:
        _, extension = server_filename.rsplit(".", 1)
    else:
        extension = None

    if extension not in POSSIBLE_VIDEO_EXTENSIONS:
        download_handling_logger.warn(
            f"The server gave filename as {server_filename!r} but the project is unsure whether this format is what it intends to use. "
            f"Hence, the downloader will be using {DEFAULT_MEDIA_EXTENSION!r}, the file format may be different or the file may end up as corrupted"
            ", in those cases, do report at project's issue tracker."
        )
        extension = DEFAULT_MEDIA_EXTENSION

    if extension == "mpd":
        expected_download_path = expected_download_path.with_suffix(f".mkv")
    else:
        expected_download_path = expected_download_path.with_suffix(f".{extension}")

    download_handling_logger.info(
        f"Server filename: {server_filename!r}, project inferred: {expected_download_path.name!r}",
    )

    expected_download_path.parent.mkdir(parents=True, exist_ok=True)

    if FFMPEG_USE_FOR_DOWNLOADS and (extension in FFMPEG_EXTENSIONS and has_ffmpeg()):
        return_code = ffmpeg_download(url, headers, expected_download_path, **opts)

        if return_code:
            raise Exception(
                f"ffmpeg exited with non-zero return code {return_code}, download failed."
            )

        return expected_download_path

    if extension in ("mpd",):
        raise NotImplementedError(
            "ffmpeg is currently disabled or not in PATH, DASH streams cannot be downloaded."
        )

    if extension in HLS_STREAM_EXTENSIONS:
        return hls_download(session, url, expected_download_path, headers or {}, **opts)

    if use_internet_download_manager:
        return idm_download(url, headers or {}, expected_download_path, **opts)

    return standard_download(
        session,
        url,
        expected_download_path,
        content_size,
        headers,
        ranges,
        log_level=opts.get("log_level", 20),
        retry_timeout=opts.get("retry_timeout", 5.0),
    )



"""
Utilities for animdl

Scraping is often quite messy and might require
a lot of complicated algorithms. Keeping these
complicated algorithms scattered across the codebase
is not good for maintainability.
"""

from . import http_caching, http_client, optopt, powertools, searching



import mimetypes
from pathlib import Path
from typing import TYPE_CHECKING

from .optopt import regexlib

if TYPE_CHECKING:
    from typing import Optional

CONTENT_DISP_RE = regexlib.compile(r'filename=(?:"(.+?)"|([^;]+))')


def guess_from_path(path: str) -> "Optional[str]":
    """
    Attempts to guess the filename from the path.
    """
    return Path(path).name


def guess_from_content_disposition(content_disposition: str) -> "Optional[str]":
    """
    Attempts to guess the filename from the content disposition header.
    """
    match = CONTENT_DISP_RE.search(content_disposition)

    if match is None:
        return None

    return guess_from_path(match.group(1) or match.group(2))


def guess_from_content_type(filename, content_type):
    """
    Attempts to guess the filename from the content type header.
    """

    guessed_extension = mimetypes.guess_extension(content_type)

    if guessed_extension is None or guessed_extension == ".bin":
        return filename

    return (filename or "file") + guessed_extension



from base64 import b64decode
from hashlib import md5

from Cryptodome.Cipher import AES
from Cryptodome.Util.Padding import unpad


def pkcs5_bytes_to_key(salt: bytes, secret, *, output=48, hash_module=md5):
    key = hash_module(secret + salt).digest()
    current_key = key

    while len(current_key) < output:
        key = hash_module(key + secret + salt).digest()
        current_key += key

    return current_key[:output]


def decipher_salted_aes(encoded_url: str, key_finders, *, aes_mode=AES.MODE_CBC):
    pwd_from_url = ""
    untouched_url = encoded_url

    for start, end in key_finders:
        pwd_from_url += untouched_url[start:end]
        encoded_url = encoded_url.replace(untouched_url[start:end], "")

    raw_value = b64decode(encoded_url.encode("utf-8"))

    assert raw_value.startswith(b"Salted__"), "Not a salt."

    key = pkcs5_bytes_to_key(raw_value[8:16], pwd_from_url.encode("utf-8"))
    return (
        unpad(AES.new(key[:32], aes_mode, key[32:]).decrypt(raw_value[16:]), 16)
        .decode("utf-8", "ignore")
        .lstrip(" ")
    )
